const path = require('path')
const fs = require('fs')
const ora = require('ora')
const merge = require('webpack-merge')
const getConfig = require('../../lib/get-config')
const { chalk, installation, logger, packageExists, resolveCwd } = require('../../utils')
const pack = require('./pack')
const zip = require('./zip')

/**
 * 获取配置
 * @param {String} source 入口文件
 * @param {String} destination 输出目录
 * @param {Object} options 配置项
 * @returns {Object}
 */
function getWebpackConfig(source, destination, options) {
  let webpackConfig = getConfig(source, destination, { ...options, mode: 'production' })

  if (options.config) {
    const filepath = resolveCwd(options.config)

    if (!fs.existsSync(filepath)) {
      logger.error(`未找到 ${filepath} 文件`)
      process.exit(2)
    }

    const customConfig = require(filepath)

    webpackConfig = merge(webpackConfig, customConfig)
  }

  return webpackConfig
}

/**
 * 运行指令
 * @param {String} source 入口文件
 * @param {String} destination 输出目录
 * @param {Object} options 配置项
 */
function run(source, destination, options) {
  return new Promise((resolve, reject) => {
    const webpack = require('webpack')

    const config = getWebpackConfig(source, destination, options)

    const spinner = ora(chalk.cyan('正在编译... \n')).start()

    webpack(config, async (err, stats) => {
      if (err) {
        spinner.fail(chalk.red('编译失败 \n'))
        logger.error(err.message)
        return reject()
      }

      if (stats.hasErrors()) {
        spinner.fail(chalk.red('编译失败 \n'))
        logger.error(stats.compilation.errors)
        return reject()
      }

      spinner.succeed(chalk.green('编译成功 \n'))

      let pkgname = ''
      try {
        if (options.zip) {
          // 编译后zip包，用于安装导入
          await zip(config.output.path, config)
        } else {
          // 编译后tgz包，用于包管理器
          pkgname = await pack(config.output.path)
        }
        resolve(pkgname)
      } catch (error) {
        reject(error)
      }
    })
  })
}

/**
 * 文件检测
 * @param {String} source 入口文件
 */
function checkFiles(source) {
  const extname = path.extname(source)
  // 目录
  const dir = extname ? path.dirname(source) : source
  // 需校验的文件
  const files = [
    { name: 'docs' },
    { name: 'package.json' },
    { name: 'config.json' }
  ]
  // 标记状态
  let flag = true

  for (let i = 0; i < files.length; i++) {
    const file = files[i]
    const filepath = path.join(resolveCwd(dir), file.name)

    if (!fs.existsSync(filepath)) {
      flag = false
      logger.error(`${file.name} 文件不存在!`)
      break
    }
  }

  return flag
}

/**
 * 编译构建
 * @param {String} source 入口文件
 * @param {String} destination 输出目录
 * @param {Object} options 配置项
 */
function build(source, destination, options) {
  if (!checkFiles(source)) {
    return
  }

  return new Promise(async (resolve, reject) => {
    // 生成的包路径
    let pkg

    if (packageExists('webpack')) {
      pkg = await run(source, destination, options)
      resolve(pkg)
    } else {
      installation('webpack', () => {
        logger.error(`检测到 ${chalk.bold('webpack')} 未安装，正在帮你进行安装`)
      }).then(async () => {
        logger.success(`${chalk.bold('webpack')} 安装成功`)
        pkg = await run(source, destination, options)
        resolve(pkg)
      }).catch(() => {
        logger.error(`执行 build 执行失败, 请再次重试 或 手动安装${chalk.bold('webpack')}`)
        reject()
      })
    }
  })
}

module.exports = build
