const path = require('path')
const fs = require('fs')
const fse = require('fs-extra')
const { chalk, logger } = require('../../utils')
const { getTimeUuid } = require('../../utils/string')
const { getPackageJson } = require('../../utils/get-package')
const compressing = require('compressing')
const pipe = require('multipipe')

/**
 * 多个文件进行压缩
 * @param {Array} filesAry ['\\xxx\\yyyy\\zzz.text', '\\nnn\\mmm\\bb.zip']
 * @param {String} name 输出路径
 * @returns true/false
 */
function getZipForFiles(filesAry, name) {
  return new Promise((resolve, reject) => {
    const zipStream = new compressing.zip.Stream()
    for (let index = 0; index < filesAry.length; index++) {
      const element = filesAry[index]
      zipStream.addEntry(element)
    }

    const destStream = fs.createWriteStream(name)

    pipe(zipStream, destStream, (err) => {
      if (err) {
        logger.error(`多文件打包压缩失败, 原因：${err}`)
        reject(err)
        return
      } else {
        resolve(true)
      }
    })
  })
}

/**
 * 文件(夹)压缩
 * @param { String } filepath 要压缩的路径
 * @param { String } name 成为压缩文件的路径
 * @param { String } type 压缩类型
 * @returns resolve reject
 */
function getZipForDirAndFile(filepath, name, type) {
  return new Promise(async (resolve, reject) => {
    if (!filepath) {
      console.log(chalk.red('压缩失败, 压缩路径为空 \n'))
      reject(false)
      return
    }

    // 必须带 .zip xxx.zip
    if (!name) {
      console.log(chalk.red('压缩失败, 压缩包名称为空 \n'))
      reject(false)
      return
    }

    if (!type) {
      console.log(chalk.red('压缩失败, 未选择压缩方式（file or dir） \n'))
      reject(false)
      return
    }

    console.log(`正在压缩 ${filepath}`)

    try {
      await compressing.zip[type](filepath, name)
      console.log(chalk.green(`文件压缩成功 ${path.basename(name)} \n`))
      resolve(true)
    } catch (error) {
      console.log(chalk.red(`压缩失败, 原因 ${error} \n`))
      reject(error)
    }
  })
}

/**
 * 读取配置文件
 * @param { string } dir 文件路径
 * @returns string / ""
 */
function getConfigData(dir) {
  const filepath = path.join(dir, 'config.json')

  if (fs.existsSync(filepath)) {
    const data = require(filepath)

    return JSON.stringify(data)
  }

  return ''
}

/**
 * 处理资源包
 * @param { String } execPath 执行命令的路径
 * @param { String } absolutePath 拼接完制造的zip文件夹路径，用来存放json npm 数据
 * @param { Object } param2 { group, component, version } 分组名 包名 版本
 * @returns Objcect / false
 */
async function zipComponent(execPath, absolutePath, { group, component, version }) {
  const dir = path.join(absolutePath, `${group}/${component}/${version}/package`)

  try {
    await fse.emptyDir(dir)
    await fse.copy(execPath, dir)

    const originalFilePath = path.join(absolutePath, `/${group ? group : component}`)
    const zipFilePath = path.join(absolutePath, '/npm.zip')

    await getZipForDirAndFile(originalFilePath, zipFilePath, 'compressDir')

    return zipFilePath
  } catch (error) {
    logger.error(error)
    return new Error('压缩组件包错误!')
  }
}

/**
 * 处理表单数据
 * @param { String } execPath 执行命令的路径
 * @param { Object } webpackConfig webpack config
 * @param { String } absolutePath 拼接完制造的zip文件夹路径，用来存放json npm 数据
 * @param { object } param3 { group, component, version } 分组名 包名 版本
 * @returns Objcect / false
 */
async function zipJson(execPath, webpackConfig, absolutePath, { group, component, version }) {
  const uploadFileBaseUrl = '/preview/component/npm'
  const comFormenginePath = `${group ? '/' + group : ''}/${component}/${version}`

  const entryKey = Object.keys(webpackConfig.entry)[0]

  const componentJsUrlLib = '/' + webpackConfig.output.filename.split('/').slice(0, -1).join('/') + `/${entryKey}.js`
  // "/preview/component/npm/pcsp/reportForm/1.2.1/package/lib/reportForm.js"
  const componentJsUrl = uploadFileBaseUrl + comFormenginePath + '/package' + componentJsUrlLib

  // "/preview/component/npm/pcsp/reportForm/1.2.1/package/config.json"
  const componentConfigUrl = uploadFileBaseUrl + comFormenginePath + '/package/config.json'

  // "/preview/component/npm/pcsp/reportForm/1.2.1/package/package.json"
  const componentPackageJsonUrl = uploadFileBaseUrl + comFormenginePath + '/package/package.json'

  // json数据
  const jsonData = {
    rid: getTimeUuid(),
    componentPrimaryKey: getTimeUuid(),
    componentName: component,
    componentVerNum: version,
    componentSetting: getConfigData(execPath),
    componentType: '0', // 写死的，代表前端
    componentGroupId: group,
    componentArtifactId: component,
    formRid: null,
    subSystemCode: null,
    subSystemName: null,
    sysId: null,
    sysTitle: null,
    sysType: null,
    componentJsUrl,
    componentConfigUrl,
    componentPackageJsonUrl,
    imagePath: null,
    componentClassRid: '',
    componentClassName: '',
    createTime: '',
    updateTime: ''
  }

  const originalFilePath = path.join(absolutePath, 'formengine.json') // ejs 输出路径\
  const zipFilePath = path.join(absolutePath, 'component.zip')

  // 开始处理 json
  try {
    await fse.writeJson(originalFilePath, jsonData)

    await getZipForDirAndFile(originalFilePath, zipFilePath, 'compressFile')

    return zipFilePath
  } catch (error) {
    logger.error(error)
    return new Error('压缩JSON文件错误!')
  }
}

/**
 * 压缩为zip类型
 * @param { String } execPath 执行命令中的路径
 * @param { Objcect } webpackConfig webpack config
 * @returns error / ""
 *
 * zip 中生成文件目录
 * dist
 *    | -- line (编译生成文件夹)
 *    | -- line-1.0.0-com (打包生成)
 *          | -- formengine.json (表单数据 formengine.zip)
 *          | -- groupName (npm.zip)
 *                         | -- name
 *                                 | -- version
 *                                         | -- package (写死的名称，里面包含编译后的资源数据)
 *                                                 | -- lib
 *                                                 | -- package.json
 *                                                 | -- xxx
 *
 * zip 完成后
 * dist
 *    | -- line
 *    | -- line-1.0.0-com
 *                | -- line-1.0.0-com.zip
*/
function zip(execPath, webpackConfig) {
  return new Promise(async (resolve, reject) => {
    // ora 跟下面的代码逻辑会冲突，控制台的打印信息会被 ora 覆盖
    // const spinner = ora(chalk.cyan('正在打包... \n')).start()
    console.log(chalk.cyan('开始 zip 压缩... \n'))

    // 数据处理
    const packageJson = getPackageJson(execPath)
    const names = packageJson.name.includes('/') ? packageJson.name.split('/').concat(packageJson.version) : ['', packageJson.name, packageJson.version]

    if (names.length > 3) {
      logger.error(`package 中 name 属性命名不规范 \n`)
      reject()
      return
    }

    // eslint-disable-next-line prefer-const
    let [group, component, version] = names
    group = group.replace(/^@/, '')
    const folder = `${group ? group + '-' : ''}` + `${component}-${version}`
    const absolutePath = path.join(path.dirname(execPath), folder)

    try {
      await fse.emptyDir(absolutePath)
    } catch (error) {
      logger.error(`创建初始文件夹失败，失败原因：${error} \n`)
      reject()
      return
    }

    Promise.all([
      zipJson(execPath, webpackConfig, absolutePath, { group, component, version }),
      zipComponent(execPath, absolutePath, { group, component, version })
    ]).then(async (files) => {
      const filename = path.join(path.dirname(execPath), `${folder}.zip`)

      await getZipForFiles(files, filename)

      // 移除文件
      fse.remove(absolutePath)

      console.log(`${chalk.green('zip 执行完成，生成压缩包')} ${chalk.yellow(filename)} \n`)

      resolve()
    }).catch(error => {
      logger.error(`压缩失败：${error} \n`)
      reject()
      return
    })
  })
}

module.exports = zip
