const path = require('path')
const fs = require('fs')
const { resolveCwd } = require('../utils')

/**
 * 新建组件
 * @param name 组件名称
 */
function create(name, options) {
  const dir = options.path ? resolveCwd(`${options.path}/${name}`) : resolveCwd(`packages/${name}`)

  // 确保文件夹存在
  fs.mkdirSync(dir, { recursive: true })

  // 获取组件模板
  const files = getTemplate(name)
  // 创建组件
  files.forEach(file => {
    fs.writeFileSync(path.join(dir, file.filename), file.content)
  })
}

/**
 * 获取组件模板
 * @param name 组件名称
 * @returns {({filename: string, content: string}[]|{filename: string, content: string}|{filename: string, content: *})[]}
 */
function getTemplate(name) {
  return [getVue(name), getJs(name), getCss(name), getPackage(name), getConfig(name)]
}

/**
 * 获取vue模板
 * @param name 组件名称
 * @returns {{filename: string, content: string}[]}
 */
function getVue(name) {
  return {
    filename: `${name}.vue`,
    content: `
<template>
  <div class="${name}"></div>
</template>

<script>
export default {
  name: '${name}'
}
</script>

<style lang='scss' scoped>
@import './index.scss';
</style>
      `.trim()
  }
}

/**
 * 获取css模板
 * @param name 组件名称
 * @returns {{filename: string, content: string}}
 */
function getCss() {
  return {
    filename: path.join('./', `index.scss`),
    content: ``
  }
}

/**
 * 获取js模板
 * @param name 组件名称
 * @returns {{filename: string, content: string}}
 */
function getJs(name) {
  return {
    filename: `${name}.js`,
    content: `
import ${name} from './${name}.vue'

export default ${name}
    `.trim()
  }
}

/**
 * 获取pkg文件模板
 * @param name 组件名称
 * @returns {{filename: string, content: *}}
 */
function getPackage(name) {
  return {
    filename: 'package.json',
    content: `
{
  "name": "${name}",
  "version": "1.0.0",
  "description": "",
  "main": "lib/${name}.js",
  "author": "",
  "license": "ISC",
  "dependencies": {}
}
    `.trim()
  }
}

/**
 * 获取config文件模板
 * @param name 组件名称
 * @returns {{filename: string, content: *}}
 */
function getConfig(name) {
  return {
    filename: 'config.json',
    content: `
{
  "name": "${name}",
  "options": {},
  "optionsConfig": []
}
    `.trim()
  }
}

module.exports = create
