const styleguidist = require('vue-styleguidist')
const ora = require('ora')
const path = require('path')
const fs = require('fs')
const getConfig = require('../lib/get-config')
const utils = require('../utils')

/**
 * 获取配置
 * @param {String} source 入口文件
 * @param {Object} options 配置项
 * @returns {Object}
 */
function getStyleguideConfig(source, options = {}) {
  const webpackConfig = getConfig(source, null, { mode: 'development' })

  const entryPath = Object.values(webpackConfig.entry)[0]

  let config = {
    styleguideDir: path.join(path.dirname(entryPath), 'docs/html'), // 文件输出位置
    usageMode: 'expand', // 文档中属性和方法的标签初始化状态，是否展开
    progressBar: true, // 打包进度条
    logger: {
      warn: console.warn,
      info: console.log,
      debug: console.log
    },
    components: entryPath,
    webpackConfig: {
      resolve: webpackConfig.resolve,
      resolveLoader: webpackConfig.resolveLoader,
      module: webpackConfig.module,
      plugins: webpackConfig.plugins
    }
  }

  if (options.config) {
    const filepath = utils.resolveCwd(options.config)

    if (!fs.existsSync(filepath)) {
      utils.logger.error(`未找到 ${filepath} 文件`)
      process.exit(2)
    }

    config = require(filepath)
  }

  return config
}

/**
 * 自动生成文档
 * @param {String} source 目标文件
 * @param {*} options
 */
function docs(source, options) {
  return new Promise((resolve, reject) => {
    const config = getStyleguideConfig(source, options)

    const spinner = ora(utils.chalk.cyan('自动生成文档... \n')).start()

    const styleguide = styleguidist.default(config)

    styleguide.build((err, config, stats) => {
      if (err || stats.hasErrors()) {
        spinner.fail(utils.chalk.red('文档生成失败 \n'))
        utils.logger.error(err || stats.compilation.errors)
        reject()
      } else {
        spinner.succeed(utils.chalk.green('文档生成成功 \n'))
        utils.logger.log(`文档已输出到 ${utils.chalk.magenta(config.styleguideDir)} \n`)
        resolve()
      }
    })
  })
}

module.exports = docs
