const path = require('path')
const fs = require('fs')
const merge = require('webpack-merge')
const webpack = require('webpack')
const WebpackDevServer = require('webpack-dev-server')
const HtmlWebpackPlugin = require('html-webpack-plugin')
const FriendlyErrorsWebpackPlugin = require('friendly-errors-webpack-plugin')
const utils = require('../utils')
const getConfig = require('../lib/get-config')

const bootPath = 'boot'

function serve(source, options = {}) {
  const components = getComponents(source)

  if (!components || components.length === 0) {
    utils.logger.error('未查询到 vue 组件!')
    return
  }

  createRender(components)

  startServer(options)
}

/**
 * 获取组件文件
 * @param {String} [source] 指定组件文件
 * @returns
 */
function getComponents(source) {
  const components = []

  if (source) { // 指定组件文件
    const componentPath = getComponentPath(source)
    if (componentPath) {
      components.push({
        name: path.basename(componentPath, path.extname(componentPath)),
        path: componentPath
      })
    }
  } else { // 不指定组件，读取所有 packages 目录下的组件
    const from = utils.resolveCwd('packages')
    if (fs.existsSync(from)) {
      const files = fs.readdirSync(from)
      files.forEach((file) => {
        const filepath = path.join(from, file)
        const stat = fs.lstatSync(filepath)
        if (stat.isDirectory()) {
          const componentPath = getComponentPath(`packages/${file}`)
          if (componentPath) {
            components.push({
              name: file,
              path: componentPath
            })
          }
        }
      })
    }
  }

  return components
}

/**
 * 获取组件的文件路径
 * @param {String} source 组件路径
 * @returns
 */
function getComponentPath(source) {
  let componentPath

  const absolutePath = utils.resolveCwd(source)
  const extname = path.extname(absolutePath)

  if (!extname) {
    const ext = ['index.vue', `${path.basename(absolutePath)}.vue`]
    for (let i = 0; i < ext.length; i++) {
      const filepath = path.join(absolutePath, ext[i])
      if (fs.existsSync(filepath)) {
        componentPath = path.join(source, ext[i])
        break
      }
    }
  } else {
    if (extname !== '.vue') {
      utils.logger.error('请指定 .vue 文件!')
      process.exit(2)
    }

    if (!fs.existsSync(absolutePath)) {
      utils.logger.error(`${absolutePath} 文件不存在!`)
      process.exit(2)
    }

    componentPath = source
  }

  return componentPath
}

/**
 * 启动服务
 * @param {Object} options
 */
function startServer(options) {
  // webpack 配置
  const webpackConfig = getWebpackConfig(options)

  // WebpackDevServer 配置
  const devServerConfig = getDevServerConfig(webpackConfig)

  let compiler

  try {
    compiler = webpack(webpackConfig)
  } catch (err) {
    if (err instanceof webpack.WebpackOptionsValidationError) {
      utils.logger.error(err.message)
      process.exit(1)
    }
    throw err
  }

  let server
  try {
    server = new WebpackDevServer(compiler, devServerConfig)
  } catch (err) {
    if (err.name === 'ValidationError') {
      utils.logger.err(err.message)
      process.exit(1)
    }
    throw err
  }

  server.listen(options.port, options.host, (err) => {
    if (err) {
      throw err
    }
  })
}

/**
 * 获取 webpack 的配置项
 */
function getWebpackConfig(options) {
  // 自定义配置
  let customConfig = {}
  if (options.config) {
    const filepath = utils.resolveCwd(options.config)

    if (!fs.existsSync(filepath)) {
      utils.logger.error(`未找到 ${filepath} 文件`)
      process.exit(2)
    }

    customConfig = require(filepath)
  }

  // 默认配置
  const defaultConfig = getConfig(`${bootPath}/main.js`, null, {
    mode: 'development',
    devtool: '#cheap-module-eval-source-map',
    output: {
      filename: 'index.js'
    }
  })

  // 合并配置
  const webpackConfig = merge(
    defaultConfig,
    {
      externals: [],
      plugins: [
        new webpack.HotModuleReplacementPlugin(),
        new FriendlyErrorsWebpackPlugin(),
        new HtmlWebpackPlugin({
          filename: 'index.html',
          template: utils.resolveCwd(`${bootPath}/index.html`),
          hash: true
        })
      ]
    },
    customConfig
  )

  return webpackConfig
}

/**
 * 获取 webpackDevServer 的配置项
 */
function getDevServerConfig(webpackConfig) {
  // 默认配置
  const devServerConfig = {
    noInfo: true, // 不在控制台打印打包的信息，有警告、错误还是会抱怨
    compress: true, // 启用 gzip 压缩
    clientLogLevel: 'none', // 当使用内联模式(inline mode)时，阻止在控制台显示消息
    overlay: { // 编译发生错误时再浏览器窗口提升
      errors: true,
      warnings: false
    },
    progress: true, // 将运行进度输出到控制台
    hot: true, // 启用热更新
    disableHostCheck: true,
    open: true, // 自动打开浏览器
    watchOptions: { // 监视文件相关的控制选项
      ignored: /node_modules/
    },
    watchContentBase: true,
    stats: webpackConfig.stats || {}
  }

  const { devServer } = merge({
    devServer: devServerConfig
  }, {
    devServer: webpackConfig.devServer
  })

  return devServer
}

/**
 * 创建渲染文件
 * @param {Array} components
 */
function createRender(components) {
  const dir = utils.resolveCwd(bootPath)

  // 确保文件夹存在
  fs.mkdirSync(dir, { recursive: true })

  const files = [getHtml(), getJs(), getVue(components)]
  // 创建组件
  files.forEach(file => {
    fs.writeFileSync(path.join(dir, file.filename), file.content)
  })
}

function getHtml() {
  return {
    filename: 'index.html',
    content: `
<!DOCTYPE html>
<html lang="en">
  <head>
    <meta charset="UTF-8">
    <meta name="viewport" content="width=device-width,initial-scale=1,shrink-to-fit=no">
    <title>组件预览</title>
  </head>
  <body>
    <div id="app"></div>
    <!-- built files will be auto injected -->
  </body>
</html>
    `.trim()
  }
}

function getJs() {
  return {
    filename: 'main.js',
    content: `
import Vue from 'vue'
import App from './App.vue'

Vue.config.productionTip = false

new Vue({
  render: h => h(App)
}).$mount('#app')
    `.trim()
  }
}

function getVue(components) {
  return {
    filename: 'App.vue',
    content: `
<template>
  <div id="app">${components.map(component => `
    <${component.name} />`).join('')}
  </div>
</template>

<script>${components.map(component => `
import ${component.name} from '../${component.path.replace(/\\/g, '/')}'`).join('')}

export default {
  name: 'App',
  components: {${components.map(component => `
    ${component.name},`).join('')}
  }
}
</script>

<style lang="scss">
</style>
    `.trim()
  }
}

module.exports = serve
