const { VueLoaderPlugin } = require('vue-loader')
const { CleanWebpackPlugin } = require('clean-webpack-plugin')
const BundleAnalyzerPlugin = require('webpack-bundle-analyzer').BundleAnalyzerPlugin
const CopyWebpackPlugin = require('copy-webpack-plugin')
const TerserPlugin = require('terser-webpack-plugin')

const path = require('path')
const fs = require('fs')
const { resolveCwd, logger } = require('../utils')
const { camelize, capitalize } = require('../utils/string')

const needCopyFiles = ['package.json', 'config.json', 'README.md', 'component.jpg', 'docs/']

/**
 * 默认配置
 */
function getDefaultConfig() {
  return {
    resolve: {
      extensions: ['.ts', '.js', '.jsx', '.vue', '.scss'],
      alias: {
        '@': resolveCwd('src')
      },
      modules: [
        'node_modules',
        resolveCwd('node_modules'),
        path.resolve(__dirname, '..', 'node_modules') // 指定当前 cli 工程目录下的 node_modules
      ]
    },
    resolveLoader: {
      modules: [
        'node_modules',
        resolveCwd('node_modules'),
        path.resolve(__dirname, '..', 'node_modules')
      ]
    },
    externals: {
      vue: 'vue'
    },
    module: {
      rules: [
        {
          test: /\.m?jsx?$/,
          exclude: [/(node_modules|bower_components)/],
          use: [
            {
              loader: 'cache-loader',
              options: {
                cacheDirectory: resolveCwd('node_modules/.cache/babel-loader')
              }
            },
            {
              loader: 'babel-loader'
            }
          ]
        },
        {
          test: /\.tsx?$/,
          use: [
            'cache-loader',
            {
              loader: 'babel-loader',
              options: {
                presets: [
                  ['@babel/preset-env']
                ],
                plugins: [
                  [
                    '@babel/transform-runtime',
                    {
                      helpers: false,
                      regenerator: true
                    }
                  ]
                ]
              }
            },
            {
              loader: 'ts-loader',
              options: {
                transpileOnly: true
              }
            }
          ]
        },
        {
          test: /\.vue$/,
          exclude: [/node_modules/],
          use: [
            {
              loader: 'cache-loader',
              options: {
                cacheDirectory: resolveCwd('node_modules/.cache/vue-loader')
              }
            },
            {
              loader: 'vue-loader',
              options: {
                prettify: false
              }
            }
          ]
        },
        {
          test: /\.(png|jpe?g|gif|webp)(\?.*)?$/,
          loader: 'url-loader'
          // options: {
          //   limit: 4096,
          //   name: 'static/img/[name].[hash:8].[ext]'
          // }
        },
        {
          test: /\.(svg)(\?.*)?$/,
          loader: 'url-loader'
        },
        {
          test: /\.(woff2?|eot|ttf|otf)(\?.*)?$/,
          loader: 'url-loader'
          // options: {
          //   limit: 4096,
          //   name: 'static/fonts/[name].[hash:8].[ext]'
          // }
        },
        {
          test: /\.(mp4|webm|ogg|mp3|wav|flac|aac)(\?.*)?$/,
          loader: 'url-loader'
          // options: {
          //   limit: 4096,
          //   name: 'static/media/[name].[hash:8].[ext]'
          // }
        },
        {
          test: /\.(sa|sc|c)ss$/,
          use: [
            'style-loader',
            'css-loader',
            {
              loader: 'postcss-loader',
              options: {
                postcssOptions: {
                  plugins: [
                    require('autoprefixer')
                  ]
                }
              }
            },
            'sass-loader'
          ]
        }
      ]
    },
    plugins: [
      new VueLoaderPlugin(),
      new CleanWebpackPlugin()
    ]
  }
}

/**
 * 获取名称
 * @param {String} entry 入口文件
 * @returns
 */
function getName(entry) {
  const dir = path.dirname(entry)
  // 文件名
  const filename = path.basename(entry, path.extname(entry))

  const packPath = path.join(resolveCwd(dir), 'package.json')

  // package.json 里 main 字段的名称
  let libname
  if (fs.existsSync(packPath)) {
    const packageJson = require(packPath)
    const main = packageJson.main
    libname = path.basename(main, path.extname(main))
  }

  const name = libname || filename

  return name
}

/**
 * 解析入口文件
 * @param {*} entry
 * @returns
 */
function getEntryPath(entry) {
  let entryPath = ''

  const absolutePath = resolveCwd(entry)
  const extname = path.extname(absolutePath)

  if (!extname) {
    const ext = ['/index.js', '/index.jsx', '/index.ts', 'index.tsx', `${path.basename(absolutePath)}.js`]
    for (let i = 0; i < ext.length; i++) {
      const filepath = path.join(absolutePath, ext[i])
      if (fs.existsSync(filepath)) {
        entryPath = filepath
        break
      }
    }
  } else {
    if (fs.existsSync(absolutePath)) {
      entryPath = absolutePath
    }
  }

  return entryPath
}

/**
 * 生成配置
 * @param {String} entry 入口文件
 * @param {String} output 输出目录
 * @param {Object} options 配置项
 * @returns
 */
function getConfig(entry, output = 'dist', options = {}) {
  const entryPath = getEntryPath(entry)

  if (!entryPath) {
    logger.error(`${entry} 入口文件不存在!`)
    process.exit(2)
  }

  const webpackConfig = getDefaultConfig()

  const name = options.name || getName(entry)
  // 配置入口
  webpackConfig.entry = {
    [name]: entryPath
  }

  // 配置出口
  webpackConfig.output = options.output || {
    filename: 'lib/[name].js',
    library: capitalize(camelize(name)),
    libraryTarget: 'umd',
    path: resolveCwd(`${output}/${name}`),
    libraryExport: 'default',
    globalObject: 'typeof self !== "undefined" ? self : this'
  }

  const isProd = options.mode === 'production'
  webpackConfig.mode = options.mode || 'none'
  webpackConfig.devtool = options.devtool || 'source-map'

  if (isProd) {
    // 打包压缩
    webpackConfig.optimization = {
      minimize: true,
      minimizer: [
        new TerserPlugin({
          parallel: true
        })
      ]
    }

    // 文件拷贝
    webpackConfig.plugins.push(
      new CopyWebpackPlugin(needCopyFiles.map(file => {
        const isFolder = /\/$/.test(file)

        return {
          from: resolveCwd(`${entryPath}/../${file}`),
          to: isFolder ? resolveCwd(`${output}/${name}/${file}`) : resolveCwd(`${output}/${name}`)
        }
      }))
    )
  }

  // 打包分析
  if (options.analyzer) {
    webpackConfig.plugins.push(
      new BundleAnalyzerPlugin({ analyzerMode: 'static' })
    )
  }

  return webpackConfig
}

module.exports = getConfig
